# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

"""The Posterity admin command

This admin command dynamically detects available commands using the
'posterity.admin_commands' entry point.
"""
from optparse import OptionParser
import os.path
import sys

from elixir import metadata, objectstore
from pkg_resources import iter_entry_points
from sqlalchemy import create_engine

from posterity.util import local


class AdminCommand(object):
    """Posterity admin command base class

    New admin commands can dynamically be added to the `posterity-admin`
    script by subclassing this class and registering it under the
    `posterity.admin_commands` entry point.
    """
    cmd_name = None
    short_help = ''
    usage = '%prog yourcommand [options] arg1 arg2 ...'

    def __init__(self, cmd_name):
        self.cmd_name = cmd_name
        self.parser = OptionParser(usage=self.usage)
        self.parser.add_option('-D', '--dburi', dest='dburi',
                               help='Database to connect to', 
                               metavar='DBURI')

    def setup_optparser(self):
        """Configure the `OptionParser`
        
        Commands that want to accept additional command line options need
        to define them using The "self.parser.add_option(...)" method.
        This method will be called before the command line is parsed
        """
        
    def finalize_options(self):
        """Decide in the final value for all options and arguments
        
        This methiod is called after the command line has been parsed and 
        the resulting options and arguments are stored in self.options and 
        self.args.
        
        If a valid database connection string is stored in self.db_url
        the following self.run()-call will be wrapped inside a database
        transaction.
        """

    def run(self):
        """Cary out the command's actions
        
        This method is invoked after the command line has been parsed.
        All `AdminCommand` subclasses need to implement this method.
        """
        raise NotImplemented

    def run_command(self, argv):
        self.setup_optparser()
        self.options, self.args = self.parser.parse_args(argv)
        self.finalize_options()
        # Should we connect to the db and start a transaction before
        # calling self.run()?
        if self.options.dburi:
            if self.options.dburi != 'test:///fixture':
                engine = create_engine(self.options.dburi)
                metadata.connect(engine)
                local.trans = objectstore.create_transaction()
            try:
                self.run()
            except:
                raise
            else:
                objectstore.flush()
                if self.options.dburi != 'test:///fixture':
                    local.trans.commit()
        else:
            return self.run()


def usage(subcommands):
    text = ('usage: %s command <options> arg1 arg2...\n'
            '\nAvailable commands:\n' % os.path.basename(sys.argv[0]))
    for cmd in subcommands:
        text += ' - %-15s %s\n' % (cmd.cmd_name, cmd.short_help)
    print >> sys.stderr, text

def run_admin(args):
    commands = {}
    for entry in iter_entry_points('posterity.admin_commands'):
        commands[entry.name] = entry.load()(entry.name)

    if len(args) < 2 or args[1] not in commands:
        usage(commands.values())
        sys.exit(1)
    cmd = commands[args[1]]
    return cmd.run_command(args[2:])

def main():
    run_admin(sys.argv)

if __name__ == '__main__':
    main()
