# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

from Cookie import SimpleCookie
from urlparse import urlparse

from elixir import objectstore, metadata
from paste.httpexceptions import HTTPSeeOther, HTTPForbidden
from paste.request import get_cookie_dict, parse_formvars
from paste.wsgilib import catch_errors
from paste.wsgiwrappers import WSGIRequest
from sqlalchemy import create_engine
import selector

from paste.urlparser import make_pkg_resources

from posterity.model import UserSession, Account
from posterity.util import local
from posterity import controllers


def posterity_app(global_config, dburi=None, base_url=None):
    """Posterity WSGI entry-point"""
    conv = controllers.ConvController()
    login = controllers.LoginController()
    msg = controllers.MsgController()
    tag = controllers.TagController()
    ident = controllers.IdentityController()
    s = selector.Selector()  
    chrome = make_pkg_resources({}, 'Posterity', 'posterity/static')
    s.add('/', GET=redirect_to_inbox)
    s.add('/chrome|', GET=chrome)
    s.add('/login', GET=login.login, POST=login.login)
    s.add('/logout', GET=login.logout)
    s.add('/tags', GET=tag.manage_form, POST=tag.create)
    s.add('/tags/{tag}', GET=tag.edit_form,
                         PUT=tag.update, DELETE=tag.delete)
    s.add('/tags/{tag}/convs', GET=conv.conv_index, POST=conv.conv_index)
    s.add('/tags/{tag}/convs/{conv}', GET=conv.conv_show, POST=conv.conv_show)
    s.add('/msgs', POST=msg.create)
    s.add('/msgs;compose_form', GET=msg.compose_form)
    s.add('/msgs/{msg};reply_form', GET=msg.reply_form)
    s.add('/identities/{ident}', GET=ident.edit_form, PUT=ident.update, 
                                 DELETE=ident.delete)
    s.add('/identities', GET=ident.manage_form, POST=ident.create)
    s.parser = lambda x: x  
    s.add(r'^/msgs/(?P<msg>\w+)/attachments/(?P<id>.+)$', 
          GET=msg.get_attachment)
    app = CSRFMiddleware(s)
    app = SessionMiddleware(app)
    app = TransactionMiddleware(app, dburi)
    if base_url:
        app = PatchEnvironMiddleware(app, base_url)
    app = FormRewriteMiddleware(app)
    return app

def redirect_to_inbox(environ, start_response):
    redir = HTTPSeeOther.relative_redirect('/tags/Inbox/convs', environ)
    return redir.wsgi_application(environ, start_response)


class PatchEnvironMiddleware(object):
    """Patches the WSGI environment according to the base_url parameter

    When running WSGI applications behind apache and mod_proxy or similar
    solutions it might be necessary to specify a base_url in order for
    the application to be able to generate correct urls.
    """
    def __init__(self, app, base_url):
        self.app = app
        o = urlparse(base_url, 'http')
        hostname = o[1]
        if ':' in hostname:
            port = o[1].split(':', 1)[1]
        elif o[0] == 'https':
            port = '443'
        else:
            port = '80'
        self.patch = {'HTTP_HOST': hostname, 'SERVER_PORT': port, 
                      'wsgi.url_scheme': o[0]}

    def __call__(self, environ, start_response):
        environ.update(self.patch)
        return self.app(environ, start_response)


class FormRewriteMiddleware(object):
    """Detect simulated PUT/DELETE requests.

    Most browsers only support form actions of GET and POST.
    `FormRewriteFilter` works around this problem by setting the action
    form attribute to "POST" and storing the real method in a hidden
    form variable.
    This middleware detects these "simulated" requests and sets the
    "REQUEST_METHOD" environ variable to the intended value.
    """
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        if environ['REQUEST_METHOD'] == 'POST':
            POST = parse_formvars(environ, include_get_vars=False)
            if '_method' in POST:
                environ['REQUEST_METHOD'] = POST['_method'].upper()
                del POST['_method']
        return self.app(environ, start_response)


class CSRFMiddleware(object):
    """WSGI filter that validates form tokens
    
    A special hidden variable (_form_token) is used to protect the 
    application against CSRF attacks. All submitted non-GET forms need
    to contain a valid form token. The form token value is a shared secret
    only known by the server and the end user.
    """
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        if environ['REQUEST_METHOD'] != 'GET' and environ['session'].account:
            form_token = environ['session'].cookie
            POST = parse_formvars(environ, include_get_vars=False)
            form_token = environ['session'].cookie
            if POST.get('_form_token') != form_token:
                forbidden = HTTPForbidden(detail='Reason: Invalid form token.')
                return forbidden(environ, start_response)
            del POST['_form_token']
        return self.app(environ, start_response)


class TransactionMiddleware(object):
    """This wsgi-app wraps each request in a database transaction.

    The transaction is rolled back if an exception is raised of if the
    HTTP status code is 5xx. Otherwise the transaction will be committed.
    """
    def __init__(self, app, dburi):
        self.app = app
        if dburi:
            engine = create_engine(dburi)
            metadata.connect(engine)

    def __call__(self, environ, start_response):
        objectstore.clear()
        local.trans = objectstore.create_transaction()
        state = dict(rollback=False, active=True, trans=local.trans)
        def _start_response(status, headers, exc_info=None):
            # We need to rollback transactions for status 5xx responses
            # even if no exception was raised.
            if status.startswith('5'):
                state['rollback'] = True
            return start_response(status, headers, exc_info)
        def close_callback():
            """close_callback might be invoked instead of ok_callback if the
            http connection is closed before the full payload has been sent"""
            if state['active']:
                objectstore.flush()
                state['trans'].commit()
        def ok_callback():
            if state['rollback']:
                state['trans'].rollback()
            else:
                objectstore.flush()
                state['trans'].commit()
            state['active'] = False
        def error_callback(exc_info):
            state['trans'].rollback()
            state['active'] == False
        app_iter = catch_errors(self.app, environ, _start_response,
                                error_callback=error_callback,
                                ok_callback=ok_callback)
        if not isinstance(app_iter, (list, tuple)):
            app_iter.close = close_callback
        return app_iter


class SessionMiddleware(object):
    """A WSGI middleware that wraps each request inside a session.
    
    Non-authenticated users are automatically redirected to the /login url.
    """
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        environ['selector'] = self.app
        cookies = get_cookie_dict(environ)
        us = None
        if 'posterity-session' in cookies:
            us = UserSession.get_by(cookie=cookies['posterity-session'])
        if not us:
            path_info = environ['PATH_INFO']
            if not path_info == '/login' and not path_info.startswith('/chrome'):
                redir = HTTPSeeOther.relative_redirect('/login', environ)
                return redir.wsgi_application(environ, start_response)
            us = UserSession()
            objectstore.expunge(us)
        environ['session'] = us
        def _start_response(status, headers, exc_info=None):
            if not status.startswith('5'):
                c = SimpleCookie()
                c['posterity-session'] = us.cookie
                c['posterity-session']['path'] = '/'
                headers.append(('Set-Cookie', c.output(header='')))
                if us.account:
                    objectstore.save(us)
            return start_response(status, headers, exc_info)
        return self.app(environ, _start_response)
