# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

import email
from time import mktime

import formencode
from paste.fileapp import DataApp
from paste.httpexceptions import HTTPNotFound
from paste.httpheaders import CONTENT_DISPOSITION

from posterity.msg import MessageRenderer
from posterity.model import *
from posterity.util import *
from posterity.web import Request, validate


class TagNameValidator(formencode.validators.UnicodeString):
    """Tag name validator. Makes sure the name is not alreay used
    """
    messages = {'exists': 'This tag is already in use'}
    def validate_python(self, value, state):
        if Tag.get_by(account=state.account, name=value):
            raise formencode.Invalid(self.message('exists', state), 
                                     value, state)


class IdentityValidator(formencode.validators.FancyValidator):
    messages = {'invalid': 'Invalid identity id'}
    def _to_python(self, value, state):
        identity = Identity.get_by(account=state.account, id=int(value))
        if not identity:
            raise formencode.Invalid(self.message('invalid', state), 
                                     value, state)
        return identity


class TagSchema(formencode.Schema):
    name = TagNameValidator(not_empty=True)


class IdentitySchema(formencode.Schema):
    name = formencode.validators.UnicodeString(not_empty=True)
    email = formencode.validators.Email(not_empty=True)
    organization = formencode.validators.UnicodeString(if_missing='')
    reply_to = formencode.validators.UnicodeString(if_missing='')
    signature = formencode.validators.UnicodeString(if_missing='')


class ComposerSchema(formencode.Schema):
    identity = IdentityValidator(not_empty=True)
    to = formencode.validators.UnicodeString(not_empty=True)
    cc = formencode.validators.UnicodeString()
    subject = formencode.validators.UnicodeString(not_empty=True)
    body = formencode.validators.UnicodeString(not_empty=True)
    in_reply_to = formencode.validators.UnicodeString()
    references = formencode.validators.UnicodeString()


class LoginController(object):

    def login(self, environ, start_response):
        req = Request(environ, start_response)
        if req.method == 'POST':
            password = encrypt_password(req.POST['password'])
            account = Account.get_by(username=req.POST['username'], 
                                     password=password)
            if account:
                environ['session'].account = account
                return req.redirect('/tags/Inbox/convs')
        return req.send_template('login.html')    

    def logout(self, environ, start_response):
        req = Request(environ, start_response)
        environ['session'].logout()
        return req.redirect('/login')


class ConvController(object):

    def perform_action(self, req, tag, do_action, sel, index_mode=False):
        """Implements common operations on a set of conversations.

        This method is used by many of the views controlled by this class.
        """
        redirect_url = '/tags/' + tag.name + '/convs'
        for s in sel:
            c = Conv.get(s)
            if do_action == 'archive':
                c.remove_tag('Inbox')
                c.mark_as_read(deleted=True)
            elif do_action == 'ham':
                c.remove_tag('Spam')
                c.undelete()
            elif do_action == 'spam':
                c.add_tag('Spam')
                c.mark_as_deleted()
                c.mark_as_read(deleted=True)
            elif do_action == 'delete':
                c.add_tag('Trash')
                c.mark_as_deleted()
                c.mark_as_read(deleted=True)
            elif do_action == 'undelete':
                c.remove_tag('Trash')
                c.undelete()
            elif do_action.startswith('tag_'):
                t = do_action[4:]
                c.add_tag(t)
                if not index_mode:
                    redirect_url = '/tags/%s/convs/%d' % (tag.name, c.id)
            elif do_action.startswith('untag_'):
                t = do_action[6:]
                c.remove_tag(t)
                if not index_mode:
                    redirect_url = '/tags/%s/convs/%d' % (tag.name, c.id)
        return req.redirect(redirect_url)

    def conv_index(self, environ, start_response):
        """The conversation index view.

        This view lists all conversations associated with a certain `tag`.
        """
        req = Request(environ, start_response)
        account = environ['session'].account
        tag = Tag.get_by(name=req.urlvars['tag'], account_id=account.id)
        if not tag:
            return HTTPNotFound()(environ, start_response)
        if req.method == 'POST':
            sel = req.POST.getall('sel')
            do_action = req.POST['do_action']
            return self.perform_action(req, tag, do_action, sel, 
                                       index_mode=True)
        offset = int(req.GET.get('offset', '0'))
        page_size = 25
        deleted = tag.name in ['Trash', 'Spam']
        convs = tag.get_convs(offset=offset, limit=page_size,
                              deleted=deleted)
        data = dict(convs=convs, tag=tag, offset=offset, page_size=page_size)
        return req.send_template('conv_index.html', data)

    def conv_show(self, environ, start_response):
        """Shows a specific conversation thread. 
    
        All messages associated with the conversation are listed below 
        each other in cronological order.
        """
        req = Request(environ, start_response)
        account = environ['session'].account
        tag = Tag.get_by(name=req.urlvars['tag'], account_id=account.id)
        conv = Conv.get(req.urlvars['conv'])
        if not conv or conv.account != account:
            return HTTPNotFound()(environ, start_response)
        if req.method == 'POST':
            do_action = req.POST['do_action']
            return self.perform_action(req, tag, do_action, [conv.id])
        # The 'Trash' and 'Spam' tags are special.
        if tag.name in ['Trash', 'Spam']:
            msgs = conv.deleted_msgs
        else:
            msgs = conv.msgs
        renderer = MessageRenderer()
        data = dict(tag=tag, conv=conv, msgs=msgs, renderer=renderer)
        return req.send_template('conv_show.html', data)


class MsgController(object):
    
    @validate(schema=ComposerSchema, on_error='/msgs;compose_form')
    def create(self, environ, start_response):
        account = environ['session'].account
        req = Request(environ, start_response)
        send_msg(account, req.formvars['identity'],
                req.formvars['subject'], 
                reflow_quoted_text(req.formvars['body']),
                req.formvars['to'], req.formvars['cc'],
                req.formvars['in_reply_to'], req.formvars['references'])
        req.flash('Message successfully sent')
        return req.redirect('/tags/Sent%20Mail/convs')

    def compose_form(self, environ, start_response):
        account = environ['session'].account
        req = Request(environ, start_response)
        if not account.identities:
            req.flash('You need to add an identity before you can send an '
                      'email')
            return req.redirect('/identities')
        data = dict(subject='', to='', cc='', body='',
                    in_reply_to='', references='')
        return req.send_template('composer.html', data)
    
    def reply_form(self, environ, start_response):
        account = environ['session'].account
        req = Request(environ, start_response)
        if not account.identities:
            req.flash('You need to add an identity before you can send an '
                      'email')
            return req.redirect('/identities')
        msg = Msg.get(req.urlvars['msg'])
        if not msg or msg.account != account:
            return HTTPNotFound()(environ, start_response)
        message = email.message_from_string(msg.payload)
        if 'Reply-To' in message:
            to = decode_header(message['Reply-To'])
        else:
            to = decode_header(message.get('From', ''))            
        cc = decode_header(message.get('Cc', ''))
        subject = decode_header(message.get('Subject', ''))
        if subject and subject[:4].lower() != 're: ':
            subject = 'Re: ' + subject
        body = msg.quote_body()
        if 'Reply-To' in message:
            references = '%s \n\t%s' % (decode_header(message['References']),
                                        decode_header(message['Message-ID']))
        else:
            references = decode_header(message['Message-ID'])
            
        data = dict(body=body, to=to, cc=cc, subject=subject,
                    in_reply_to=msg.msg_id, references=references)
        return req.send_template('composer.html', data)

    def get_attachment(self, environ, start_response):
        account = environ['session'].account
        req = Request(environ, start_response)
        msg = Msg.get(req.urlvars['msg'])
        if not msg or msg.account != account:
            return HTTPNotFound()(environ, start_response)
        message = email.message_from_string(msg.payload)
        part_id = req.urlvars['id']
        for part in message.walk():
            content_id = part.get('Content-Id', '').strip('<>')
            filename = part.get_filename()
            if part_id in (content_id, filename):
                last_modified = mktime(msg.date.utctimetuple())
                content_type = part.get_content_type()
                app = DataApp('', content_type=content_type)
                app.set_content(part.get_payload(decode=True),
                                last_modified=last_modified)
                CONTENT_DISPOSITION.update(app.headers, attachment=True,
                                           filename=filename)
                return app(environ, start_response)
        return HTTPNotFound()(environ, start_response)


class TagController(object):

    def delete(self, environ, start_response):
        """Handles DELETE requests for /tags/{tag}"""
        req = Request(environ, start_response)
        account = environ['session'].account
        tag = Tag.get_by(name=req.urlvars['tag'], account_id=account.id)
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        assert tag.is_special == False
        objectstore.delete(tag)
        req.flash('Tag deleted')
        return req.redirect('/tags')

    @validate(schema=TagSchema, on_error='/tags')
    def create(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        Tag(account=account, name=req.formvars['name'])
        req.flash('Tag added')
        return req.redirect('/tags')
            
    @validate(schema=TagSchema, on_error='/tags/{tag};edit_form')
    def update(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        tag = Tag.get_by(account=account, name=req.urlvars['tag'])
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        assert tag.is_special == False
        tag.name = req.formvars['name']
        req.flash('Tag updated')
        return req.redirect('/tags')

    def edit_form(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        tag = Tag.get_by(account=account, name=req.urlvars['tag'])
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        data = dict(tag=tag)
        return req.send_template('tag_edit.html', data)

    def manage_form(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        
        conn = local.trans.connection(Conv)
        s = select([conv_to_tag.c.tag_id, func.count(conv_to_tag.c.conv_date),
                    func.count(conv_to_tag.c.conv_date2)],
                   and_(Tag.c.account_id==account.id,
                        Tag.c.id==conv_to_tag.c.tag_id),
                   group_by=conv_to_tag.c.tag_id)
        conv_count = {}
        for tag_id, count_normal, count_deleted in conn.execute(s):
            conv_count[tag_id] = count_normal
        data = dict(account=account, conv_count=conv_count)
        return req.send_template('tags.html', data)


class IdentityController(object):
    
    def manage_form(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        data = dict()
        return req.send_template('identity_list.html', data)

    def edit_form(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        ident = Identity.get(req.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        data = dict(ident=ident)
        return req.send_template('identity_edit.html', data)

    @validate(schema=IdentitySchema, on_error='/identities')
    def create(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        Identity(account=account, **req.formvars)
        req.flash('Identity added')
        return req.redirect('/identities')
            
    @validate(schema=IdentitySchema, on_error='/identities/{ident}')
    def update(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        ident = Identity.get(req.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        ident.name = req.formvars['name']
        ident.email = req.formvars['email']
        ident.reply_to = req.formvars['reply_to']
        ident.organization = req.formvars['organization']
        ident.signature = req.formvars['signature']
        req.flash('Identity updated')
        return req.redirect('/identities')

    def delete(self, environ, start_response):
        req = Request(environ, start_response)
        account = environ['session'].account
        ident = Identity.get(req.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        objectstore.delete(ident)
        req.flash('Identity deleted')
        return req.redirect('/identities')
